/*******************************************************************************
* FILE NAME: user_routines.c <VEX VERSION>
*
* DESCRIPTION:
*  This file contains the default mappings for CONFIGURATIONS A,B and C.  
*
* USAGE:
*  You can either modify this file to fit your needs, or remove it from your 
*  project and replace it with a modified copy. 
*
*******************************************************************************/

#include <stdio.h>
#include <math.h>
#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "user_routines.h"
#include "auto_routines.h"

#define CODE_VERSION            2     /* Use this def. to keep track of your version # */

#define BUTTON_REV_THRESH       100
#define BUTTON_FWD_THRESH       154
#define NEUTRAL_VALUE           127

void Steer_1(void);
void Steer_2(void);
void Steer_3(void);


/*** DEFINE USER VARIABLES AND INITIALIZE THEM HERE ***/
/* EXAMPLES: (see MPLAB C18 User's Guide, p.9 for all types)
unsigned char wheel_revolutions = 0; (can vary from 0 to 255)
unsigned int  delay_count = 7;       (can vary from 0 to 65,535)
int           angle_deviation = 142; (can vary from -32,768 to 32,767)
unsigned long very_big_counter = 0;  (can vary from 0 to 4,294,967,295)
*/


/*******************************************************************************
* FUNCTION NAME: Limit_Switch_Max
* PURPOSE:       Sets a PWM value to neutral (127) if it exceeds 127 and the
*                limit switch is on.
* CALLED FROM:   this file
* ARGUMENTS:     
*     Argument       Type             IO   Description
*     --------       -------------    --   -----------
*     switch_state   unsigned char    I    limit switch state
*     *input_value   pointer           O   points to PWM byte value to be limited
* RETURNS:       void
*******************************************************************************/
void Limit_Switch_Max(unsigned char switch_state, unsigned char *input_value)
{
  if (switch_state == CLOSED)
  { 
    if(*input_value > 127)
      *input_value = 127;
  }
}


/*******************************************************************************
* FUNCTION NAME: Limit_Switch_Min
* PURPOSE:       Sets a PWM value to neutral (127) if it's less than 127 and the
*                limit switch is on.
* CALLED FROM:   this file
* ARGUMENTS:     
*     Argument       Type             IO   Description
*     --------       -------------    --   -----------
*     switch_state   unsigned char    I    limit switch state
*     *input_value   pointer           O   points to PWM byte value to be limited
* RETURNS:       void
*******************************************************************************/
void Limit_Switch_Min(unsigned char switch_state, unsigned char *input_value)
{
  if (switch_state == CLOSED)
  { 
    if(*input_value < 127)
      *input_value = 127;
  }
}


/*******************************************************************************
* FUNCTION NAME: Limit_Mix
* PURPOSE:       Limits the mixed value for one joystick drive.
* CALLED FROM:   Default_Routine, this file
* ARGUMENTS:     
*     Argument             Type    IO   Description
*     --------             ----    --   -----------
*     intermediate_value    int    I    
* RETURNS:       unsigned char
*******************************************************************************/
unsigned char Limit_Mix (int intermediate_value)
{
  static int limited_value;
  
  if (intermediate_value < 2000)
  {
    limited_value = 2000;
  }
  else if (intermediate_value > 2254)
  {
    limited_value = 2254;
  }
  else
  {
    limited_value = intermediate_value;
  }
  return (unsigned char) (limited_value - 2000);
}


/*******************************************************************************
* FUNCTION NAME: Setup_Who_Controls_Pwms
* PURPOSE:       Each parameter specifies what processor will control the pwm.  
*                 
* CALLED FROM:   User_Initialization
*     Argument             Type    IO   Description
*     --------             ----    --   -----------
*     pwmSpec1              int     I   USER/MASTER (defined in ifi_aliases.h)
*     pwmSpec2              int     I   USER/MASTER
*     pwmSpec3              int     I   USER/MASTER
*     pwmSpec4              int     I   USER/MASTER
*     pwmSpec5              int     I   USER/MASTER
*     pwmSpec6              int     I   USER/MASTER
*     pwmSpec7              int     I   USER/MASTER
*     pwmSpec8              int     I   USER/MASTER
* RETURNS:       void
*******************************************************************************/
static void Setup_Who_Controls_Pwms(int pwmSpec1,int pwmSpec2,int pwmSpec3,int pwmSpec4,
                                    int pwmSpec5,int pwmSpec6,int pwmSpec7,int pwmSpec8)
{
  txdata.pwm_mask = 0xFF;         /* Default to master controlling all PWMs. */
  if (pwmSpec1 == USER)           /* If User controls PWM1 then clear bit0. */
    txdata.pwm_mask &= 0xFE;      /* same as txdata.pwm_mask = txdata.pwm_mask & 0xFE; */
  if (pwmSpec2 == USER)           /* If User controls PWM2 then clear bit1. */
    txdata.pwm_mask &= 0xFD;
  if (pwmSpec3 == USER)           /* If User controls PWM3 then clear bit2. */
    txdata.pwm_mask &= 0xFB;
  if (pwmSpec4 == USER)           /* If User controls PWM4 then clear bit3. */
    txdata.pwm_mask &= 0xF7;
  if (pwmSpec5 == USER)           /* If User controls PWM5 then clear bit4. */
    txdata.pwm_mask &= 0xEF;
  if (pwmSpec6 == USER)           /* If User controls PWM6 then clear bit5. */
    txdata.pwm_mask &= 0xDF;
  if (pwmSpec7 == USER)           /* If User controls PWM7 then clear bit6. */
    txdata.pwm_mask &= 0xBF;
  if (pwmSpec8 == USER)           /* If User controls PWM8 then clear bit7. */
    txdata.pwm_mask &= 0x7F;
}


/*******************************************************************************
* FUNCTION NAME: User_Initialization
* PURPOSE:       This routine is called first (and only once) in the Main function.  
*                You may modify and add to this function.
*                The primary purpose is to set up the DIGITAL IN/OUT - ANALOG IN
*                pins as analog inputs, digital inputs, and digital outputs.
* CALLED FROM:   main.c
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
void User_Initialization (void)
{
/* FIRST: Set up the pins you want to use as analog INPUTs. */
  IO1 = IO2 = INPUT;        /* Used for analog inputs. */
    /* 
     Note: IO1 = IO2 = IO3 = IO4 = INPUT 
           is the same as the following:

           IO1 = INPUT;
           IO2 = INPUT;
           IO3 = INPUT;
           IO4 = INPUT;
    */

/* SECOND: Configure the number of analog channels. */
  Set_Number_of_Analog_Channels(TWO_ANALOG);     /* See ifi_aliases.h */

/* THIRD: Set up any extra digital inputs. */
  /* The six INTERRUPTS are already digital inputs. */
  /* If you need more then set them up here. */
  /* IOxx = IOyy = INPUT; */
  IO3 = IO4 = IO5 = IO6 = IO7 = IO8 = INPUT;        
  IO9 = IO10 = IO11 = IO12 = IO13 = IO15 = INPUT;        

/* FOURTH: Set up the pins you want to use as digital OUTPUTs. */
  IO14 = IO16 = OUTPUT;      

/* FIFTH: Initialize the values on the digital outputs. */
  rc_dig_out14 = rc_dig_out16 = 0; 

/* SIXTH: Set your initial PWM values.  Neutral is 127. */
  pwm01 = pwm02 = pwm03 = pwm04 = pwm05 = pwm06 = pwm07 = pwm08 = 127;

/* SEVENTH: Choose which processor will control which PWM outputs. */
  Setup_Who_Controls_Pwms(MASTER,MASTER,MASTER,MASTER,MASTER,MASTER,MASTER,MASTER);

/* EIGHTH: Set your PWM output type.  Only applies if USER controls PWM 1, 2, 3, or 4. */
  /*   Choose from these parameters for PWM 1-4 respectively:                          */
  /*     IFI_PWM  - Standard IFI PWM output generated with Generate_Pwms(...)          */
  /*     USER_CCP - User can use PWM pin as digital I/O or CCP pin.                    */
  Setup_PWM_Output_Type(IFI_PWM,IFI_PWM,IFI_PWM,IFI_PWM);

  /* 
     Example: The following would generate a 40KHz PWM with a 50% duty cycle
              on the CCP2 pin (PWM OUT 1):

      Setup_Who_Controls_Pwms(USER,USER,MASTER,MASTER,MASTER,MASTER,MASTER,MASTER);
      CCP2CON = 0x3C;
      PR2 = 0xF9;
      CCPR2L = 0x7F;
      T2CON = 0;
      T2CONbits.TMR2ON = 1;
      Setup_PWM_Output_Type(USER_CCP,IFI_PWM,IFI_PWM,IFI_PWM);
  */

/* Add any other user initialization code here. */

  Initialize_Serial_Comms();     

  Putdata(&txdata);             /* DO NOT CHANGE! */
  User_Proc_Is_Ready();         /* DO NOT CHANGE! - last line of User_Initialization */

#ifdef _SIMULATOR
    statusflag.NEW_SPI_DATA = 1;
#else
  /* This code receives the 1st packet from master to obtain the version # */
  while (!statusflag.NEW_SPI_DATA);  /* Wait for 1st packet from master */
  Getdata(&rxdata);   
  printf("VEX - Master v%d, User v%d\r",(int)rxdata.master_version,(int)CODE_VERSION);        
#endif
}


/*******************************************************************************
* FUNCTION NAME: Process_Data_From_Master_uP
* PURPOSE:       Executes every 17ms when it gets new data from the master 
*                microprocessor.
* CALLED FROM:   main.c
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
void Process_Data_From_Master_uP(void)
{
	static unsigned int i = 0;
	static unsigned int j = 0;
	static unsigned int k = 0;
	static long accum = 0;
	unsigned long temp_accum;
	int temp_gyro_rate;
	int temp_gyro_angle;
	int temp_gyro_bias;

  Getdata(&rxdata);   /* Get fresh data from the master microprocessor. */

  Default_Routine();  /* Optional.  See below. */

  /* Add your own code here. */

  Putdata(&txdata);             /* DO NOT CHANGE! */
}


/*******************************************************************************
* FUNCTION NAME: Default_Routine
* PURPOSE:       Performs the default mappings of inputs to outputs for  
*                CONFIGURATIONS A, B, and C (B = drive mode 12 - settings on radio)
* CALLED FROM:   this file, Process_Data_From_Master_uP routine
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
#define ALLOWED_SERVO_RANGE  255
#define DEADBAND 10
void Default_Routine(void)
{
	int servo_direction;

//printf("PWM_in1=%d, in2=%d, in3=%d, in4=%d\r",PWM_in1,PWM_in2,PWM_in3,PWM_in4);
if ((PWM_in3<=(127+DEADBAND)) && (PWM_in3>=(127-DEADBAND)) &&
   (PWM_in4<=(127+DEADBAND)) && (PWM_in4>=(127-DEADBAND)))  // One joystick overrides the other!
{   // Right joystick is driving the robot
	Steer_1(); // x-axis = direction & y-axis = speed
//	Steer_2(); // Direction of joystick = direction & magnitude = speed (robot point of view)
//	Steer_3(); // Direction of joystick = direction & magnitude = speed (driver point of view)
}
// ************ Left  joystick is for spins and curved driving and overrides the right joystick
else
{
	//Use PWM_in4 for both spin speed and direction
      if (PWM_in4 < (127-DEADBAND) || PWM_in4 > (127+DEADBAND))  // Outside of deadband
	{
		pwm01 = PWM_in4;    		// LF
		pwm02 = PWM_in4;			// LR
		pwm03 = PWM_in4;			// RR
		pwm04 = PWM_in4;			// RF
//		printf("SPIN: 01-02=%d, 03-04=%d, ",pwm01, pwm03);

//	Use to check the servo range
//     if (PWM_in4 > 127) PWM_in4 = 255 - PWM_in4;	// Only allow one servo direction
//	servo_direction = (int)PWM_in4 * ALLOWED_SERVO_RANGE / 255;	

	// Automatically position wheels for full spin if any in4 input
	servo_direction = 44; 		// Optimum spin angle
	pwm05 = servo_direction;									// LF
	pwm06 = ALLOWED_SERVO_RANGE - servo_direction;	// LR
	pwm07 = servo_direction;									// RR
	pwm08 = ALLOWED_SERVO_RANGE - servo_direction;	// RF
//	printf("servo=%d, PWM_in4=%d, pwm05-7=%d, pwm06-8=%d\r",servo_direction,(int)PWM_in4,(int)pwm05,(int)pwm06);
	}
}

} /* END Default_Routine(); */


/*******************************************************************************
* FUNCTION NAME: Steer_1
* PURPOSE:       Maps OI controls to movement
*				x-axis is the direction; y-axis is the speed
*******************************************************************************/
void Steer_1(void)
{
	int servo_direction;

	//PWM_in2 is throttle or speed (this forces the robot to be slower during angled movement)
	pwm01 = PWM_in2;    		// LF
	pwm02 = PWM_in2;			// LR
	pwm03 = 255 - PWM_in2;	// RR
	pwm04 = 255 - PWM_in2;	// RF
	printf("DIR: Speed--PWM_in2=%d, ",PWM_in2);

	//PWM_in1 is steering direction
 	// Map the joystick range to the permitted driving servo range. (joystick/255 =servo/allowed range)
	servo_direction = 255 - (int)PWM_in1 * ALLOWED_SERVO_RANGE / 255;
	pwm05 = servo_direction;	// LF
	pwm06 = servo_direction;	// LR
	pwm07 = servo_direction;	// RR
	pwm08 = servo_direction;	// RF
	printf("servo=%d, PWM_in1=%d\r",servo_direction,(int)PWM_in1);
}

/*******************************************************************************
* FUNCTION NAME: Steer_2
* PURPOSE:       Maps OI controls to movement
*				Position of the joystick indicates both speed and direction
*******************************************************************************/
void Steer_2(void)
{
	int servo_direction, radius;
	float slope, x, y;

	// Standard circle equation x2 + y2 = radius2 where radius is the speed 0-127
	// sqrt of (x2 + y2) is throttle or speed (this forces the robot to be slower during angled movement)
	x = (float)PWM_in2 - 127; // Make 0 neutral (-127 to 127)
	y = (float)PWM_in1 - 127;
	radius = sqrt((x * x) + (y * y));
//\	radius += 127; // Return to 127 neutral
	pwm01 = radius;    		// LF
	pwm02 = radius;			// LR
	pwm03 = 255 - radius;	// RR
	pwm04 = 255 - radius;	// RF
	printf("DIR: PWM_in1=%d,  PWM_in2=%d, radius=%d, ", PWM_in1, PWM_in2, radius);

	// whereas, y/x = slope from joystick neutral, indicates steering direction
 	// Map the joystick range to the permitted driving servo range. (joystick/255 =servo/allowed range)
	slope = (float)y / (float)x;
	servo_direction = 255 * slope;
	pwm05 = servo_direction;	// LF
	pwm06 = servo_direction;	// LR
	pwm07 = servo_direction;	// RR
	pwm08 = servo_direction;	// RF
	printf("servo=%d \r",(int)servo_direction);

	pwm01 = pwm02 = pwm03 = pwm04 = 127;
	pwm05 = pwm06 = pwm07 = pwm08 = 127;
}



/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
